/*
 * Copyright (C) 2025 Stefan Hansson
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * Author: Stefan Hansson <newbyte@postmarketos.org>
 */

#pragma once

#include <adwaita.h>

char *ms_tweaks_util_gdkrgba_to_rgb_hex_string (const GdkRGBA *from);
