/*
 * Copyright 2023 Gotam Gorabh <gautamy672@gmail.com>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <adwaita.h>

G_BEGIN_DECLS

char *ms_generate_debug_info (void);

G_END_DECLS
