/*
 * Copyright (C) 2025 Stefan Hansson
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * Author: Stefan Hansson <newbyte@postmarketos.org>
 */
#pragma once

#include <glib.h>

G_BEGIN_DECLS

void ms_internal_init (void);
void ms_internal_uninit (void);

G_END_DECLS
